# -*- coding: utf-8 -*-
"""
Copyright © 2017 SDElements Inc.

This application/module/plugin functions as a certmonger helper and a certbot
plugin. The operation depends on the entry point.

TBD Explain different entry points and flows

https://pagure.io/certmonger/blob/master/f/doc/helpers.txt
Add documentation for certbot plugins
"""

import certbot
import certbot.errors
import certbot.plugins
import certbot.plugins.dns_common
import certbot.interfaces
import dns
import dns.name
import gssapi
import hashlib
import logging
import os
import ipalib
from ipalib.install.kinit import kinit_keytab

__version__ = '0.0.6'

# Set up logging
logger = logging.getLogger('cerlet')


class Authenticator(certbot.plugins.dns_common.DNSAuthenticator):
    """ FreeIPA / Red Hat Enterprise Linux IdM authentication using DNS challenges """

    description = __doc__.strip().split("\n", 1)[0]

    def __init__(self, *args, **kwargs):
        # Maps provided domains to IPA zones
        self.zone_map = {}
        super(Authenticator, self).__init__(*args, **kwargs)

    def more_info(self):
        return self.description

    def _setup_credentials(self):
        # Set up IPA API connection
        logger.info('Setting up IPA Connection using kerberos credentials')

        try:
            ipalib.api.bootstrap_with_global_options(context=__name__)
        except ipalib.errors.KerberosError:
            logger.exception('Exception occurred communicating and authenticating to the IPA server, make sure you have valid kerberos credentials')
            raise certbot.errors.MisconfigurationError()

        useccache = getattr(self.config, 'use_ccache_creds') if hasattr(self.config, 'use_ccache_creds') else None
        workdir = getattr(self.config, 'work_dir') if hasattr(self.config, 'work_dir') else None
        ccache = None

        if useccache:
            if not "KRB5CCNAME" in os.environ:
                raise certbot.errors.MisconfigurationError()
        else:
            keytab = getattr(self.config, 'keytab_file') if hasattr(self.config, 'keytab_file') else None
            principal = getattr(self.config, 'submitter-principal') if hasattr(self.config, 'submitter-principal') else None

            if not keytab:
                keytab = '/etc/krb5.keytab'

            if not principal:
                principal = str('host/%s' % (ipalib.api.env.host))

            os.environ['KRB5CCNAME'] = "MEMORY:"

            # Kerberos initialization
            logger.debug('Kerberos principal: %s', principal)
            ccache_filename = "MEMORY:"

            try:
                kinit_keytab(principal, keytab, ccache_filename, attempts=5)
                os.environ['KRB5CCNAME'] = ccache_filename
            except gssapi.exceptions.GSSError as e:
                sys.exit("Failed to obtain host TGT: %s" % e)
            except Exception as ex:
                logger.critical("Kerberos authentication failed: %s", ex)
                # signal failure and let init system to restart the daemon
                raise certbot.errors.MisconfigurationError()

        try:
            ipalib.api.finalize()
        except ipalib.errors.KerberosError:
            logger.exception('Exception occurred communicating and authenticating to the IPA server, make sure you have valid kerberos credentials')
            raise certbot.errors.MisconfigurationError()

        if ipalib.api.env.in_server:
            ipalib.api.Backend.ldap2.connect()
        else:
            ipalib.api.Backend.rpcclient.connect()

    def _perform(self, domain, validation_domain_name, validation):  # pragma: no cover
        logger.debug('Adding DNS Records {0} to zone: {1} with value: {2}'.format(validation_domain_name, domain, validation))
        dns_zone_candidate = dns.name.from_text(domain)
        while True:
            try:
                ipalib.api.Command.dnszone_show(idnsname=str(dns_zone_candidate.to_text()))
                record = dns.name.from_text(validation_domain_name).relativize(dns_zone_candidate)
                try:
                    ipalib.api.Command.dnsrecord_add(dnszoneidnsname=str(dns_zone_candidate.to_text()), idnsname=str(record), txtrecord=validation)
                except ipalib.errors.EmptyModlist:
                    logger.warning('DNS entry already exists: {0} with value: {1}'.format(record, validation))
                self.zone_map[validation_domain_name] = (dns_zone_candidate, record)
                break
            except (ipalib.errors.NotFound, ipalib.errors.ConversionError):
                logger.debug('Unable to find Zone: {0}, checking for parent'.format(dns_zone_candidate.to_text(omit_final_dot=True)))
                try:
                    dns_zone_candidate = dns_zone_candidate.parent()
                except dns.name.NoParent:
                    message = 'Unable to find DNS Zone on IPA server'
                    logger.exception(message)
                    raise ipalib.errors.NotFound(format=None, message=message.decode())

    def _cleanup(self, domain, validation_domain_name, validation):
        logger.debug('Removing DNS entry: {0} to zone: {1} with value: {2}'.format(validation_domain_name, domain, validation))
        ipalib.api.Command.dnsrecord_del(dnszoneidnsname=str(self.zone_map[validation_domain_name][0].to_text()), idnsname=str(self.zone_map[validation_domain_name][1]), txtrecord=validation)

    @classmethod
    def add_parser_arguments(cls, add, default_propagation_seconds=10):
        super(Authenticator, cls).add_parser_arguments(add, default_propagation_seconds)
        add('-h', '--host_name', '--host', dest='ipa_host', help='Hostname or IP Address of the IPA (Satellite) server')
        add('-d', '--ipa_domain_name', '--ipa_domain', dest='ipa_domain', help='Domain in IPA (Satellite) to register under')
        add('-H', '--xml_rpc_url', '--xmlrpc-url', dest='ipa_xml_rpc_url', help='XML RPC service location')
        add('-C', '--ca_path', '--capath', dest='ca_path', help='Path do a directory containing PEM encoded CA files')
        add('-c', '--ca_file', '--cafile', dest='ca_file', help='Path do a file containing PEM encoded CA certificates')
        add('-t', '--keytab_name', '--keytab-name', dest='keytab_file', help='Path to a keytab file containing credentials for IPA server authentication')
        add('-k', '--submitter_principal', '--submitter-principal', dest='submitter_principal', help='Kerberos principal for IPA server authentication')
        add('-K', '--use_ccache_creds', '--use-ccache-creds', dest='use_ccache_creds', default=False, action='store_true', help='Use default ccache for authorization instead of authenticating')
        add('-P', '--request_principal', '--principal-of-request', dest='request_principal', help='Principal(s) (FQDN) used in signing request, comma separated')
        add('-T', '--request_profile', '--profile', dest='request_profile', help='Use a specific profile when requesting enrollment')
