# -*- coding: utf-8 -*-
"""
Copyright © 2017 SDElements Inc.

This application/module/plugin functions as a certmonger helper and a certbot
plugin. The operation depends on the entry point.

TBD Explain different entry points and flows

https://pagure.io/certmonger/blob/master/f/doc/helpers.txt
Add documentation for certbot plugins
"""

import acme
import argparse
import certbot
import certbot._internal.account
import certbot._internal.client
import certbot._internal.constants
from certbot._internal.display import obj as display_obj
from certbot._internal.display import util as internal_display_util
import certbot._internal.plugins
import certbot.errors
import certbot.main
import certbot.interfaces
import certbot.util
import hashlib
import logging
import os
import OpenSSL
import sys
from cerlet.util import *

__version__ = '0.0.6'

# Set up logging
logger = logging.getLogger('cerlet')

class CertMongerAction(object):
    """ Represents an action requested by Certmonger """
    # Supported Exit Codes
    EXIT_ISSUED = 0
    EXIT_WAIT = 1
    EXIT_REJECTED = 2
    EXIT_UNREACHABLE = 3
    EXIT_UNCONFIGURED = 4
    EXIT_WAIT_WITH_DELAY = 5
    EXIT_OPERATION_NOT_SUPPORTED = 6
    EXIT_NEW_KEYPAIR_REQUIRED = 17

    PATHS = {'config_dir': '/etc/certmonger/letsencrypt',
             'work_dir': '/var/lib/certmonger/letsencrypt',
             'log_dir': '/var/log/letsencrypt'}

    defaults = certbot._internal.constants.CLI_DEFAULTS

    def __init__(self,
                 paths=PATHS,
                 ):

        verbosity=logging.NONE

        parser = argparse.ArgumentParser()
        parser.add_argument('-t', '--test', dest='staging', default=True, action='store_true', help='Use staging server')
        parser.add_argument('-s', '--server', dest='server', default=None, help='API Server')
        parser.add_argument('-d', '--debug', dest='debug', default=False, action='store_true', help='Enable debugging.')
        parser.add_argument('-n', '--no-verify-ssl', dest='verify_ssl', default=True, action='store_false')
        parser.add_argument('-v', '--verbose', dest='verbose_count', default=0, action='count')
        parser.add_argument('-k', "--key-size", dest="key_size", default=4096)
        parser.add_argument('-e', "--email", dest="email")
        parser.add_argument('-U', "--user-agent", dest="user_agent", default='{0}/{1}'.format(__name__, __version__))
        parser.add_argument('-p', "--propagation-seconds", dest="propagation_seconds", default=10)
        parser.add_argument(None, "--verbose-level", dest="verbose_level", default=None, help=argparse.SUPPRESS)

        args = parser.parse_args()

        if args.verbose_count is not None:
            verbosity = logging.WARNING - int(args.verbose_level) * 10
        else:
            verbosity = logging.WARNING - args.verbose_count * 10

        formatter = logging.Formatter('%(module)s.%(funcName)s: %(message)s')
        logger.setLevel(verbosity)

        # Set up logging to use syslog
        syslog_handler = logging.handlers.SysLogHandler(address='/dev/log')
        syslog_handler.setFormatter(formatter)
        syslog_handler.setLevel(verbosity)
        logger.addHandler(syslog_handler)
        # Create console logger for stdout/stderr if we have a tty, full debug
        if sys.stdout.isatty():
            logger.warning('This is tty')
            console_handler = logging.StreamHandler()
            console_handler.setLevel(logging.DEBUG)
            console_handler.setFormatter(formatter)
            logger.addHandler(console_handler)


        # Load and store relevant environment variables
        self.environment = self.load_environment_variables()

        if args.server is not None:
            self.defaults['server'] = args.server

        self.defaults['verbose_level'] = args.verbose_level
        self.defaults['verbose_count'] = args.verbose_count
        self.defaults['debug'] = args.debug

        # If no email is specified we try to get one from the request
        if not email:
            try:
                email = self.environment['CERTMONGER_REQ_EMAIL']
            except KeyError:
                email = None
                self.defaults['register_unsafely_without_email'] = True

        self.defaults['email'] = args.email

        # Set timeout for DNS propagation
        #settattr(self.defaults, 'cerlet:ipa_propagation_seconds', propagation_seconds)
        self.defaults['ipa_propagation_seconds'] = args.propagation_seconds

        # Set key size
        self.defaults['key_size'] = args.key_size

        # Configure SSL/TLS Server cert verification
        self.defaults['verify_ssl'] = args.verify_ssl

        # Set prod/staging server selector
        if args.staging:
            self.defaults['staging'] = args.staging

        # Find all plugins and set
        self.defaults['plugins'] = certbot._internal.plugins.disco.PluginsRegistry.find_all()

        # Set User Agent
        self.defaults['user_agent'] = args.user_agent

        # Override account directory
        self.defaults['accounts_dir'] = os.path.join(paths['config_dir'],
                                                     certbot._internal.constants.ACCOUNTS_DIR,
                                                     self.environment['CERTMONGER_REQ_HOSTNAME'])

        # Create any directories which don't exist with correct
        # permisisons/owner/group and set in config
        for key, path in paths.items():
            mkdirp(path, permission_mode=0o700)
            self.defaults[key] = path

        # Create config object from defaults in certbot and assign defaults
        self.namespace = argparse.Namespace(**self.defaults)
        self.config = certbot.configuration.NamespaceConfig(namespace=self.namespace)
        logger.debug('Certbot config: {}'.format(self.namespace))

        # Configure displayer depending on if we have a tty or not
        if sys.stdout.isatty():
            self.displayer = display_obj.NoninteractiveDisplay(sys.stdout)
        else:
            self.displayer = display_obj.NoninteractiveDisplay(open(os.devnull, "w"))

        display_obj.set_display(self.displayer)

        # Set up Certbot Account Storage, at some point this should be moved
        # and stored by IPA in LDAP. For now we create separate accounts per
        # host and store details on the filesystem.
        account_storage = certbot._internal.account.AccountFileStorage(self.config)
        accounts = account_storage.find_all()

        # Assume there will only be one account per host and that it will never
        # need to be updated with a new email address ...
        if len(accounts) == 0:
            # Register account with Let's Encrypt Server if needed, we always agree
            # to the TOS terms (see lambda).
            account, acme = certbot._internal.client.register(config=self.config,
                                                    account_storage=account_storage,
                                                    tos_cb=lambda *_, **__: True)

            account_storage.save(account, acme)
        else:
            acme = None
            account = accounts[0]

        # Instantiate authenticator using DNS/FreeIPA
        self.plugin = self.defaults['plugins']['ipa'].init(config=self.config)
        # Instantiate the installer using Certmonger
        #self.installer = certbot._internal.plugins.disco.PluginsRegistry.find_all()['ipa']

        # Instantiate client
        self.client = certbot._internal.client.Client(config=self.config,
                                            account_=account,
                                            auth=self.plugin,
                                            installer=self.plugin,
                                            acme=acme)


    @staticmethod
    def _raise_not_implemented(exception=NotImplementedError):
        """ Raises the provided exception """
        raise exception

    @staticmethod
    def load_environment_variables(pattern='CERTMONGER'):
        """ Loads environment variables matching a pattern """
        matches = {}
        for key, value in os.environ.items():
            if pattern in key:
                matches[key] = value

        return matches

    @classmethod
    def operation_factory(cls, operation):
        """
        Returns the appopriate method for an operation as they are
        specified by CertMonger in environment variables.
        """
        logger.debug('Certmonger operation detected, evaluating environment variables and performing actions as requested')
        valid_operations = {'SUBMIT': cls.submit,
                            'POLL': cls.poll,
                            'IDENTIFY': cls.identify,
                            'GET-NEW-REQUEST-REQUIREMENTS': cls.requirements,
                            'GET-RENEW-REQUEST-REQUIREMENTS': cls.renew_requirements,
                            'GET-SUPPORTED-TEMPLATES': cls.templates,
                            'GET-DEFAULT-TEMPLATE': cls.default_template,
                            'FETCH-SCEP-CA-CAPS': cls._raise_not_implemented,
                            'FETCH-SCEP-CA-CERTS': cls._raise_not_implemented,
                            'FETCH-ROOTS': cls.get_ca_root_certs}

        logger.error('Running opration: {0}'.format(operation))

        try:
            return valid_operations[operation](cls())
        except NotImplementedError:
            logger.warning('Operation {0} is not supported in this version of {1} ({2})'.format(operation, __name__, __version__))
            return cls.EXIT_OPERATION_NOT_SUPPORTED
        except KeyError:
            logger.error('Unknown Operation: {0}'.format(operation))
            return cls.EXIT_OPERATION_NOT_SUPPORTED

    def register(self, config=None):
        certbot.main.register(config or self.config, unused_plugins=None)

    def submit(self, csr=None, domains=None, ca_profile=None, ca_nickname=None, ca_issuer=None):
        """
        Accepts a single Certificate Signing Request, PKCS#10/PEM encoded

        First time enrollment requested by Certmonger

        Keyword arguments:

        csr -- Enrollment request in PEM form. PKCS#10 format, PEM encoded
        ca_profile -- Name of enrollment profile/template/certtype
        ca_nickname --  Name by which the CA is known, and would have been specified to the -c
            option to the "getcert" command. For example "prod" or "test"
        ca_issuer --  Requested issuer for enrollment

        Issues certificates will be returned in PEM encoded x.509 format
        and an exit status of 0 will be returned.

        If the certificates can't be issued immediately a "cookie" value/token
        will be returned and an exit status of 1 returned. The requestor is
        free to try again at their convenience presenting the cookie/token.

        If the request is rejected an error will returned along with an exit
        status of 2.

        If there was a connection or networking exception the error will be
        returned along with an exit status of 3.

        If additional data is required the specific error will be returned
        along with an exit status of 4.


        * If the client should wait for a specific period of time (for example, if
            the CA has told it when to try again), output a delay size in seconds, a
            newline, and a "cookie" value, and exit with status 5.  The daemon will try
            again after the specified amount of time has passed.

        * If the CA indicates that the client needs to try again using a different
            key pair in the signing request (for example, if its policy limits the
            number of times a particular key pair can be enrolled, or the length of
            time one can be in service), exit with status 17.  The daemon will generate
            a new key pair and try again.

        * If the helper does not understand what is being asked of it, exit with
            status 6.  You should never return this value for "SUBMIT" or "POLL", but
            it is mentioned here so that we can refer to this list later.
        """
        logger.debug('Submitting certificate signing request')
        mycsr = certbot.util.CSR(data = csr or self.environment['CERTMONGER_CSR'], form = 'pem', file = None )
        domains = domains or [self.environment['CERTMONGER_REQ_SUBJECT']]

        try:
            cert, chain = self.client.obtain_certificate_from_csr(mycsr)
            sys.stdout.write(cert.decode('utf-8'))

            logger.debug('Certificate signing is complete.')
            logger.debug('Certificate: {}'.format(cert))
            logger.debug('Chain: {}'.format(chain))

            return self.EXIT_ISSUED

        except acme.messages.Error as e:
            logger.error('acme.messages.Error')
            logger.error(e)
            sys.stdout.write(str(e))
            return self.EXIT_REJECTED

        except acme.error.ClientError as e:
            logger.error('acme.error.ClientError')
            logger.error(e)
            sys.stdout.write(str(e))
            return self.EXIT_UNREACHABLE

        except (optparse.BadOptionError, Exception) as e: # Catch all for exceptions
            logger.error('Exception')
            logger.exception(e)
            sys.stdout.write(str(e))
            return self.EXIT_REJECTED

    def poll(self, cookie=None):
        """
        Poll status of previously submitted request

        Keyword arguments:

        cookie -- A cookie/token to identify a previously submitted request

        If the submit method previously returned with status 1 or 5, this can
        be called to retry.

        Returned certificates and exit statuses are the same as the submit
        method.
        """
        logger.debug('Polling for certificate signing request status')
        cookie = cookie or self.environment['CERTMONGER_CA_COOKIE']

    @staticmethod
    def identify():
        """
        Outputs the version of the helper and returns an exit status of 0
        """
        logger.debug('Returning version of helper/plugin')
        return __version__

    def requirements(self, renew=False):
        """
        Returns a list of required arguments/environment variables for the
        poll and submit methods, if renew is true it lists those required when
        renewing a certificate rather than requesting a new one.
        """
        submit_requirements = 'CERTMONGER_CSR,CERTMONGER_REQ_SUBJECT,CERTMONGER_OPERATION,CERTMONGER_REQ_HOSTNAME,CERTMONGER_REQ_EMAIL'
        logger.debug('Returning list of required attributes/arguments:')
        logger.debug(submit_requirements)
        sys.stdout.write(submit_requirements)

    def renew_requirements(self):
        return self.requirements(renew=True)

    def templates(default_only=False):
        """
        Returns list of all profiles/templates/cert-types, by default all
        supported templates are returned.
        """
        logger.debug('Returning templates/profiles')

    def default_template(self):
        return self.templates(default_only=True)

    def get_ca_root_certs(self):
        """
        Return a dictionary of nickname/cert with the cert in PEM format.
        """
        logger.debug('Dumping CA Root certificates')
        cert_path = os.path.join(os.path.split(__file__)[0], 'certs')
        root_certs = {'dstrootx3.pem': b'06:87:26:03:31:A7:24:03:D9:09:F1:05:E6:9B:CF:0D:32:E1:BD:24:93:FF:C6:D9:20:6D:11:BC:D6:77:07:39',
                      'isrgrootx1.pem': '96:BC:EC:06:26:49:76:F3:74:60:77:9A:CF:28:C5:A7:CF:E8:A3:C0:AA:E1:1A:8F:FC:EE:05:C0:BD:DF:08:C6',
                      'letsencryptauthorityx3.pem': b'73:1D:3D:9C:FA:A0:61:48:7A:1D:71:44:5A:42:F6:7D:F0:AF:CA:2A:6C:2D:2F:98:FF:7B:3C:E1:12:B1:F5:68',
                      'lets-encrypt-x3-cross-signed.pem': b'25:84:7D:66:8E:B4:F0:4F:DD:40:B1:2B:6B:07:40:C5:67:DA:7D:02:43:08:EB:6C:2C:96:FE:41:D9:DE:21:8D'}

        for filename, sha256sum in root_certs.iteritems():
            cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, open(os.path.join(cert_path, filename)).read())
            if cert.digest('sha256') == sha256sum:  # Verify integrity
                nickname = ' '.join(x[1] for x in cert.get_subject().get_components())
                if cert.get_issuer() == cert.get_subject():  # Root CA's can be identified by always being self signed
                    sys.stdout.write('\n' + nickname + '\n')
                    sys.stdout.write(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, cert))
                    sys.stdout.write('\n')
                else:
                    sys.stdout.write(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, cert))
                    sys.stdout.write('\n')
            else:
                logger.error('Unable to verify the integrity of Root CA {0}'.format(filename))
                logger.error('SHA256 mismatch {0} - {1}'.format(sha256sum, cert.digest('sha256')))
                raise IntegrityError

